#include <assert.h>
#include "pass1.h"

/*
 * We define location operations which operate on the expression tree
 * during the first pass (before sending to the backend for code generation.)
 */

/*
 * This function is called from many places in the tree building
 * code, its purpose is to let the target do early
 * transformations on the trees built.  There are a bunch
 * of target-specific stuff that can be done here but there 
 * are at least two things that must be done:
 *	Convert NAME nodes to stack references or whatever
 *	data it targets. 
 *	Cast constants to correct type.
 */
NODE * clocal(NODE *p)
{
  /* the major essential job is rewriting the
     automatic variables and arguments in terms of
     REG and OREG nodes */
  /* conversion ops which are not necessary are also clobbered here */
  /* in addition, any special features (such as rewriting
     exclusive or) are easily handled here as well */
  struct symtab *sp;
  NODE *l, *r;
  int op;

  op = p->n_op;
  sp = p->n_sp;
  l  = p->n_left;
  r  = p->n_right;

  switch( op ){
  case NAME:
    if ((sp = p->n_sp) == NULL || blevel == 0)
      return p; /* Nothing to care about */
    switch (sp->sclass) {
    case PARAM:
    case AUTO:
      /* fake up a structure reference */
      r = block(REG, NIL, NIL, PTR+STRTY, 0, 0);
      r->n_lval = 0;
      r->n_rval = FPREG;
      p = stref(block(STREF, r, p, 0, 0, 0));
      break;
    case STATIC:
      if (sp->slevel == 0)
	break;
      p->n_lval = 0;
      p->n_sp = sp;
      break;      
    case REGISTER:
      p->n_op = REG;
      p->n_lval = 0;
      p->n_rval = sp->soffset;
      break;      
    }
    break;
  case PMCONV:
  case PVCONV:
    if( p->n_right->n_op != ICON ) cerror( "bad conversion", 0);
    nfree(p);
    return(buildtree(op == PMCONV ? MUL : DIV, p->n_left, p->n_right));    
  case PCONV:
    if (l->n_op == SCONV)
      break;    
    if (l->n_op == ICON || (ISPTR(p->n_type) && ISPTR(l->n_type))) {
      l->n_type = p->n_type;
      l->n_qual = p->n_qual;
      l->n_df = p->n_df;
      l->n_sue = p->n_sue;
      nfree(p);
      p = l;
    }
    break;
  case SCONV:
    if (DEUNSIGN(p->n_type) == INT && DEUNSIGN(l->n_type) == INT) {
      nfree(p);
      p = l;
      break;
    }
    if (l->n_op != ICON)
      break;
    if (ISPTR(p->n_type)) {
      l->n_type = p->n_type;
      nfree(p);
      p = l;
      break;
    }
    switch (p->n_type) {
    case BOOL:      l->n_lval = (l->n_lval != 0); break;
    case CHAR:      l->n_lval = (char)l->n_lval; break;
    case UCHAR:     l->n_lval = l->n_lval & 0377; break;
    case SHORT:     l->n_lval = (short)l->n_lval; break;
    case USHORT:    l->n_lval = l->n_lval & 0177777; break;
    case UNSIGNED:  l->n_lval = l->n_lval & 0xffffffff; break;
    case INT:       l->n_lval = (int)l->n_lval; break;
    case ULONG:
    case ULONGLONG: l->n_lval = l->n_lval; break;
    case LONG:
    case LONGLONG:	l->n_lval = (long long)l->n_lval; break;
    case FLOAT:
    case DOUBLE:
    case LDOUBLE:
      l->n_op = FCON;
      l->n_dcon = l->n_lval;
      break;
    case VOID:
      break;
    default:
      cerror("sconv type unknown %d", p->n_type);
    }    
    l->n_type = p->n_type;
    nfree(p);
    p = l;
    break;
  case FORCE:
    /* Put attached value into the return register. */
    p->n_op = ASSIGN;
    p->n_right = p->n_left;
    p->n_left = block(REG, NIL, NIL, p->n_type, 0, MKSUE(INT));
    p->n_left->n_rval = RETREG(p->n_type);
    break;
  }  
  return p;
}

/*
 * This function is called for each node just before the
 * node tree is written out to pass2. Common usage here
 * may be to for example save floating constants to memory
 * and converting the constants to memory references.
 */
/* cc/ccom/trees.c:2017 */
void myp2tree(NODE *p)
{
  struct symtab *sp;
  if (p->n_op != FCON)
    return;
#define IALLOC(sz)	(isinlining ? permalloc(sz) : tmpalloc(sz))  
  sp = IALLOC(sizeof(struct symtab));
  sp->sclass = STATIC;
  sp->ssue = MKSUE(p->n_type);
  sp->slevel = 1; /* fake numeric label */
  sp->soffset = getlab();
  sp->sflags = 0;
  sp->stype = p->n_type;
  sp->squal = (CON >> TSHIFT);
  
  defloc(sp);
  ninval(0, sp->ssue->suesize, p);

  p->n_op = NAME;
  p->n_lval = 0;	
  p->n_sp = sp;
}

/*
 * Tell whether a node can have it's address referenced.
 * Usually called to see if ADDROF NAME nodes may be
 * converted to an ICON.  This function can be useful 
 * when dealing with PIC code.
 */
/* cc/ccom/optim.c:119 */
int andable(NODE *p)
{
  return (1);
}

/*
 * Called directly after all arguments in the header of
 * a function is declared, before bfcode().  Usually the
 * target don't need to do anything here.
 */
/* cc/ccom/pftn.c:560 */
void cendarg()
{
  autooff = AUTOINIT;
}

/*
 * Return 1 if a variable of type t is OK to put in register.
 */
/* cc/ccom/pftn.c:1126 */
int cisreg(TWORD t)
{
  if (t == FLOAT || t == DOUBLE || t == LDOUBLE)
    return 0; /* not yet */
  return 1;
}

/*
 * Return a node, for structure references, which is suitable
 * for being added to a pointer of type t, in order to be off
 * bits offset into a structure.
 * t, d, and s are the type, dimension offset, and size offset.
 * On byte-addressed targets the function can usually just
 * return a integer constant node, see i386, but on
 * word-addressed targets it may be more difficult.  See pdp10
 * for an example.
 */
/* cc/ccom/trees.c:865 */
NODE * offcon(OFFSZ off, TWORD t, union dimfun *d, struct suedef *sue)
{
  return bcon(off/SZCHAR);
}

/*
 * Generate code to allocate off bits on the stack. 
 * p is a tree that when evaluated is the multiply count
 * for off, t is a storeable node where to write the allocated
 * address.  Called for dynamic elements on the stack 
 * like alloca().
 */
/* cc/ccom/pftn.c:1259 */
void spalloc(NODE *t, NODE *p, OFFSZ off)
{
}

/*
 * Print out a string of characters stored in the sname field
 * in the symtab.  The string is stored as a null-terminated
 * string with control characters escaped.  The environment
 * must be setup by this function, for example by calling 
 * defloc().
 */
/* cc/ccom/pftn.c:1084 */
void instring(struct symtab *sp)
{
}

/*
 * The same as above but for wide strings.
 */
/* cc/ccom/pftn.c:1082 */
void inwstring(struct symtab *sp)
{
}
/*
 * These routines are used in initializations of aggregates.
 * zbits() output fsz bits of zero starting at off in the
 * current aggregate.  infld() initializes a bitfield at
 * offset off and size fsz with the constant val.
 * ninval() will do the same as infld() but the value written
 * out is associated with a node; for example an address.
 * Before any of these routines are called defloc() is called
 * to setup the environment.
 */
/* cc/ccom/init.c:653 */
void zbits(OFFSZ off, int fsz)
{
}
/* cc/ccom/init.c:72 */
void infld(CONSZ off, int fsz, CONSZ val)
{
}
/* cc/ccom/init.c:74 */
void ninval(CONSZ off, int fsz, NODE *p)
{
}

/*
 * Returns p converted to an external name, for example
 * prefixed by "_".  Memory for this may be allocated by
 * tmpalloc() or permalloc(), the latter if the compiler
 * currently is reading an inline function.
 */
/* cc/ccom/stabs.c:201 */
char *exname(char *p)
{
  return (p == NULL ? "" : p);
}

/*
 * Map a type to another type if it is not supported on the
 * local machine.  For example INT and LONG are commonly the
 * same on 32-bit machines.
 */
/* cc/ccom/scan.c */
TWORD ctype(TWORD type)
{
  /*
  switch (BTYPE(type)) {
  case SHORT:
    MODTYPE(type,INT);
    break;
  case USHORT:
    MODTYPE(type,UNSIGNED);
    break;
  case LONGLONG:
    MODTYPE(type,INT);
    break;
  case ULONGLONG:
    MODTYPE(type,UINT);
    break;
    }*/
  return (type);
}

/*
 * Called when a function call is found and read in, to let		
 * the target do something if needed.  Usually not used.
 */
/* cc/ccom/pftn.c:2059 */
void calldec(NODE *p, NODE *q) 
{
}

/*
 * Let the target declare an external referenced variable.
 * Usually not used.
 */
/* cc/ccom/pftn.c:1375 */
void extdec(struct symtab *q)
{
}

/* 
 * Define a common zeroed area for sp. (usually .comm/.lcomm).
 */
/* cc/ccom/pftn.c:1381 */
void defzero(struct symtab *sp)
{
  int off;
  off = tsize(sp->stype, sp->sdf, sp->ssue);
  off = (off+(SZCHAR-1))/SZCHAR;
  printf("        .%scomm ", sp->sclass == STATIC ? "l" : "");
  if (sp->slevel == 0)
    printf("%s,0%o\n", exname(sp->soname), off);
  else
    printf(LABFMT ",0%o\n", sp->soffset, off);
}
/*
 * Give the target the opportunity to handle a pragma 
 * statement.  ary is a tokenized array of identifiers,
 * separated by space or tab.
 */
/* cc/ccom/scan.c:2863 */
int mypragma(char **ary)
{
  return 0;
}

/*
 * Called after a variable is declared to allow for the
 * target to do something with it.
 */
/* cc/ccom/pftn.c:411 */
void fixdef(struct symtab *sp)
{
}

